@@ -232,7 +232,7 @@ static char *xstrsignal(n)
 {
   static char buf[sizeof("Signal ") + 1 + sizeof(int)*3];
 #ifdef HAVE_SYS_SIGLIST
-  extern char *sys_siglist[];
+  extern const char * const sys_siglist[];
   if (n >= 0 && n < NSIG && sys_siglist[n] != 0)
     return sys_siglist[n];
 #endif /* HAVE_SYS_SIGLIST */
