@@ -29,6 +29,8 @@ Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 #include "div.h"
 #include "reg.h"
 #include "charinfo.h"
+#include "searchpath.h"
+#include "macropath.h"
 #include <math.h>
 
 symbol default_family("T");
@@ -175,11 +177,9 @@ void environment::mark_last_line()
 
 void widow_control_request()
 {
-  if (has_arg()) {
-    int n;
-    if (get_integer(&n))
-      curenv->widow_control = n != 0;
-  }
+  int n;
+  if (has_arg() && get_integer(&n))
+    curenv->widow_control = n != 0;
   else
     curenv->widow_control = 1;
   skip_line();
@@ -553,6 +553,7 @@ environment::environment(symbol nm)
   tabs(units_per_inch/2, TAB_LEFT),
   current_tab(TAB_NONE),
   current_field(0),
+  margin_character_flags(0),
   margin_character_node(0),
   margin_character_distance(points_to_units(10)),
   numbering_nodes(0),
@@ -627,6 +628,7 @@ environment::environment(const environment *e)
   tabs(e->tabs),
   current_tab(TAB_NONE),
   current_field(0),
+  margin_character_flags(e->margin_character_flags),
   margin_character_node(e->margin_character_node),
   margin_character_distance(e->margin_character_distance),
   numbering_nodes(0),
@@ -948,33 +950,13 @@ void family_change()
   skip_line();
 }
 
-#if 0
 void point_size()
 {
   int n;
-  if (has_arg()) {
-    if (get_number(&n, 0, curenv->get_requested_point_size()/sizescale)) {
-      n *= sizescale;
-      if (n <= 0)
-	n = 1;
-      curenv->set_size(n);
-    }
-  }
-  else
-    curenv->set_size(0);
-  skip_line();
-}
-#endif
-
-void point_size()
-{
-  int n;
-  if (has_arg()) {
-    if (get_number(&n, 'z', curenv->get_requested_point_size())) {
-      if (n <= 0)
-	n = 1;
-      curenv->set_size(n);
-    }
+  if (has_arg() && get_number(&n, 'z', curenv->get_requested_point_size())) {
+    if (n <= 0)
+      n = 1;
+    curenv->set_size(n);
   }
   else
     curenv->set_size(0);
@@ -986,10 +968,8 @@ void space_size()
   int n;
   if (get_integer(&n)) {
     curenv->space_size = n;
-    if (has_arg()) {
-      if (get_integer(&n))
-	curenv->sentence_space_size = n;
-    }
+    if (has_arg() && get_integer(&n))
+      curenv->sentence_space_size = n;
     else
       curenv->sentence_space_size = curenv->space_size;
   }
@@ -1021,6 +1001,8 @@ void center()
   int n;
   if (!has_arg() || !get_integer(&n))
     n = 1;
+  else if (n < 0)
+    n = 0;
   while (!tok.newline() && !tok.eof())
     tok.next();
   if (break_flag)
@@ -1035,6 +1017,8 @@ void right_justify()
   int n;
   if (!has_arg() || !get_integer(&n))
     n = 1;
+  else if (n < 0)
+    n = 0;
   while (!tok.newline() && !tok.eof())
     tok.next();
   if (break_flag)
@@ -1047,102 +1031,85 @@ void right_justify()
 void line_length()
 {
   hunits temp;
-  if (!has_arg()) {
-    hunits temp = curenv->line_length;
-    curenv->line_length = curenv->prev_line_length;
-    curenv->prev_line_length = temp;
-  }
-  else if (get_hunits(&temp, 'm', curenv->line_length)) {
+  if (has_arg() && get_hunits(&temp, 'm', curenv->line_length)) {
     if (temp < H0) {
       warning(WARN_RANGE, "bad line length %1u", temp.to_units());
       temp = H0;
     }
-    curenv->prev_line_length = curenv->line_length;
-    curenv->line_length = temp;
   }
+  else
+    temp = curenv->prev_line_length;
+  curenv->prev_line_length = curenv->line_length;
+  curenv->line_length = temp;
   skip_line();
 }
 
 void title_length()
 {
   hunits temp;
-  if (!has_arg()) {
-    hunits temp = curenv->title_length;
-    curenv->title_length = curenv->prev_title_length;
-    curenv->prev_title_length = temp;
-  }
-  else if (get_hunits(&temp, 'm', curenv->title_length)) {
+  if (has_arg() && get_hunits(&temp, 'm', curenv->title_length)) {
     if (temp < H0) {
       warning(WARN_RANGE, "bad title length %1u", temp.to_units());
       temp = H0;
     }
-    curenv->prev_title_length = curenv->title_length;
-    curenv->title_length = temp;
   }
+  else
+    temp = curenv->prev_title_length;
+  curenv->prev_title_length = curenv->title_length;
+  curenv->title_length = temp;
   skip_line();
 }
 
 void vertical_spacing()
 {
-  if (!has_arg()) {
-    vunits temp = curenv->vertical_spacing;
-    curenv->vertical_spacing = curenv->prev_vertical_spacing;
-    curenv->prev_vertical_spacing = temp;
-  }
-  else {
-    vunits temp;
-    if (get_vunits(&temp, 'p', curenv->vertical_spacing)) {
-      if (temp <= V0) {
-	warning(WARN_RANGE, "vertical spacing must be greater than 0");
-	temp = vresolution;
-      }
-      curenv->prev_vertical_spacing = curenv->vertical_spacing;
-      curenv->vertical_spacing = temp;
+  vunits temp;
+  if (has_arg() && get_vunits(&temp, 'p', curenv->vertical_spacing)) {
+    if (temp <= V0) {
+      warning(WARN_RANGE, "vertical spacing must be greater than 0");
+      temp = vresolution;
     }
   }
+  else
+    temp = curenv->prev_vertical_spacing;
+  curenv->prev_vertical_spacing = curenv->vertical_spacing;
+  curenv->vertical_spacing = temp;
   skip_line();
 }
 
 void line_spacing()
 {
   int temp;
-  if (!has_arg()) {
-    temp = curenv->line_spacing;
-    curenv->line_spacing = curenv->prev_line_spacing;
-    curenv->prev_line_spacing = temp;
-  }
-  else if (get_integer(&temp)) {
+  if (has_arg() && get_integer(&temp)) {
     if (temp < 1) {
       warning(WARN_RANGE, "value %1 out of range: interpreted as 1", temp);
       temp = 1;
     }
-    curenv->prev_line_spacing = curenv->line_spacing;
-    curenv->line_spacing = temp;
   }
+  else
+    temp = curenv->prev_line_spacing;
+  curenv->prev_line_spacing = curenv->line_spacing;
+  curenv->line_spacing = temp;
   skip_line();
 }
 
 void indent()
 {
   hunits temp;
-  int err = 0;
-  if (!has_arg())
+  if (has_arg() && get_hunits(&temp, 'm', curenv->indent)) {
+    if (temp < H0) {
+      warning(WARN_RANGE, "indent cannot be negative");
+      temp = H0;
+    }
+  }
+  else
     temp = curenv->prev_indent;
-  else if (!get_hunits(&temp, 'm', curenv->indent))
-    err = 1;
   while (!tok.newline() && !tok.eof())
     tok.next();
   if (break_flag)
     curenv->do_break();
-  if (temp < H0) {
-    warning(WARN_RANGE, "indent cannot be negative");
-    temp = H0;
-  }
-  if (!err) {
-    curenv->have_temporary_indent = 0;
-    curenv->prev_indent = curenv->indent;
-    curenv->indent = temp;
-  }
+  curenv->have_temporary_indent = 0;
+  curenv->prev_indent = curenv->indent;
+  curenv->indent = temp;
   tok.next();
 }
 
@@ -1169,11 +1136,9 @@ void temporary_indent()
 
 void underline()
 {
-  int n = 0;
-  if (!has_arg())
+  int n;
+  if (!has_arg() || !get_integer(&n))
     n = 1;
-  else if (!get_integer(&n))
-    n = 0;
   if (n <= 0) {
     if (curenv->underline_lines > 0) {
       curenv->prev_fontno = curenv->fontno;
@@ -1189,42 +1154,51 @@ void underline()
   skip_line();
 }
 
-
 void control_char()
 {
-  if (!has_arg())
-    curenv->control_char = '.';
-  else if (tok.ch() == 0)
-    error("bad control character");
-  else
-    curenv->control_char = tok.ch();
+  curenv->control_char = '.';
+  if (has_arg()) {
+    if (tok.ch() == 0)
+      error("bad control character");
+    else
+      curenv->control_char = tok.ch();
+  }
   skip_line();
 }
 
 void no_break_control_char()
 {
-  if (!has_arg())
-    curenv->no_break_control_char = '\'';
-  else if (tok.ch() == 0)
-    error("bad control character");
-  else
-    curenv->no_break_control_char = tok.ch();
+  curenv->no_break_control_char = '\'';
+  if (has_arg()) {
+    if (tok.ch() == 0)
+      error("bad control character");
+    else
+      curenv->no_break_control_char = tok.ch();
+  }
   skip_line();
 }
 
-
 void margin_character()
 {
-  if (curenv->margin_character_node) {
-    delete curenv->margin_character_node;
-    curenv->margin_character_node = 0;
-  }
   charinfo *ci = get_optional_char();
   if (ci) {
-    curenv->margin_character_node = curenv->make_char_node(ci);
-    hunits d;
-    if (curenv->margin_character_node && has_arg() && get_hunits(&d, 'm'))
-      curenv->margin_character_distance = d;
+    node *nd = curenv->make_char_node(ci);
+    if (nd) {
+      delete curenv->margin_character_node;
+      curenv->margin_character_node = nd;
+      curenv->margin_character_flags = (MARGIN_CHARACTER_ON
+					|MARGIN_CHARACTER_NEXT);
+      hunits d;
+      if (has_arg() && get_hunits(&d, 'm'))
+	curenv->margin_character_distance = d;
+    }
+  }
+  else {
+    curenv->margin_character_flags &= ~MARGIN_CHARACTER_ON;
+    if (curenv->margin_character_flags == 0) {
+      delete curenv->margin_character_node;
+      curenv->margin_character_node = 0;
+    }
   }
   skip_line();
 }
@@ -1233,13 +1207,7 @@ void number_lines()
 {
   delete_node_list(curenv->numbering_nodes);
   curenv->numbering_nodes = 0;
-  int n;
-  if (has_arg() && get_integer(&n, next_line_number)) {
-    next_line_number = n;
-    if (next_line_number < 0) {
-      warning(WARN_RANGE, "negative line number");
-      next_line_number = 0;
-    }
+  if (has_arg()) {
     node *nd = 0;
     for (int i = '9'; i >= '0'; i--) {
       node *tem = make_node(charset_table[i], curenv);
@@ -1252,6 +1220,19 @@ void number_lines()
     }
     curenv->numbering_nodes = nd;
     curenv->line_number_digit_width = env_digit_width(curenv);
+    int n;
+    if (!tok.delimiter()) {
+      if (get_integer(&n, next_line_number)) {
+	next_line_number = n;
+	if (next_line_number < 0) {
+	  warning(WARN_RANGE, "negative line number");
+	  next_line_number = 0;
+	}
+      }
+    }
+    else
+      while (!tok.space() && !tok.newline() && !tok.eof())
+	tok.next();
     if (has_arg()) {
       if (!tok.delimiter()) {
 	if (get_integer(&n)) {
@@ -1283,11 +1264,9 @@ void number_lines()
 
 void no_number()
 {
-  if (has_arg()) {
-    int n;
-    if (get_integer(&n))
-      curenv->no_number_count = n > 0 ? n : 0;
-  }
+  int n;
+  if (has_arg() && get_integer(&n))
+    curenv->no_number_count = n > 0 ? n : 0;
   else
     curenv->no_number_count = 1;
   skip_line();
@@ -1302,10 +1281,8 @@ void no_hyphenate()
 void hyphenate_request()
 {
   int n;
-  if (has_arg()) {
-    if (get_integer(&n))
-      curenv->hyphenation_flags = n;
-  }
+  if (has_arg() && get_integer(&n))
+    curenv->hyphenation_flags = n;
   else
     curenv->hyphenation_flags = 1;
   skip_line();
@@ -1319,11 +1296,9 @@ void hyphen_char()
 
 void hyphen_line_max_request()
 {
-  if (has_arg()) {
-    int n;
-    if (get_integer(&n))
-      curenv->hyphen_line_max = n;
-  }
+  int n;
+  if (has_arg() && get_integer(&n))
+    curenv->hyphen_line_max = n;
   else
     curenv->hyphen_line_max = -1;
   skip_line();
@@ -1406,13 +1381,20 @@ void environment::newline()
 void environment::output_line(node *n, hunits width)
 {
   prev_text_length = width;
-  if (margin_character_node) {
+  if (margin_character_flags) {
     hunits d = line_length + margin_character_distance - saved_indent - width;
     if (d > 0) {
       n = new hmotion_node(d, n);
       width += d;
     }
-    node *tem = margin_character_node->copy();
+    margin_character_flags &= ~MARGIN_CHARACTER_NEXT;
+    node *tem;
+    if (!margin_character_flags) {
+      tem = margin_character_node;
+      margin_character_node = 0;
+    }
+    else
+      tem = margin_character_node->copy();
     tem->next = n;
     n = tem;
     width += tem->width();
@@ -1808,7 +1790,7 @@ void environment::final_break()
 
 void environment::do_break()
 {
-  if (curdiv == topdiv && !topdiv->first_page_begun) {
+  if (curdiv == topdiv && topdiv->before_first_page) {
     topdiv->begin_page();
     return;
   }
@@ -1867,7 +1849,7 @@ void break_request()
 
 void title()
 {
-  if (curdiv == topdiv && !topdiv->first_page_begun) {
+  if (curdiv == topdiv && topdiv->before_first_page) {
     handle_initial_title();
     return;
   }
@@ -1922,9 +1904,8 @@ void title()
 
 void adjust()
 {
-  if (!has_arg())
-    curenv->adjust_mode |= 1;
-  else
+  curenv->adjust_mode |= 1;
+  if (has_arg()) {
     switch (tok.ch()) {
     case 'l':
       curenv->adjust_mode = ADJUST_LEFT;
@@ -1942,12 +1923,17 @@ void adjust()
     default:
       int n;
       if (get_integer(&n)) {
-	if (0 <= n && n <= 5)
-	  curenv->adjust_mode = n;
-	else
+	if (n < 0)
+	  warning(WARN_RANGE, "negative adjustment mode");
+	else if (n > 5) {
+	  curenv->adjust_mode = 5;
 	  warning(WARN_RANGE, "adjustment mode `%1' out of range", n);
+	}
+	else
+	  curenv->adjust_mode = n;
       }
     }
+  }
   skip_line();
 }
 
@@ -1959,12 +1945,12 @@ void no_adjust()
 
 void input_trap()
 {
+  curenv->input_trap_count = 0;
   int n;
-  if (!has_arg())
-    curenv->input_trap_count = 0;
-  else if (get_integer(&n)) {
+  if (has_arg() && get_integer(&n)) {
     if (n <= 0)
-      error("number of lines for input trap must be greater than zero");
+      warning(WARN_RANGE,
+	      "number of lines for input trap must be greater than zero");
     else {
       symbol s = get_name(1);
       if (!s.is_null()) {
@@ -2720,12 +2706,68 @@ void init_env_requests()
 
 // Hyphenation - TeX's hyphenation algorithm with a less fancy implementation.
 
-const int WORD_MAX = 1024;
+struct trie_node;
 
-dictionary exception_dictionary(501);
+class trie {
+  trie_node *tp;
+  virtual void do_match(int len, void *val) = 0;
+  virtual void do_delete(void *) = 0;
+  void delete_trie_node(trie_node *);
+public:
+  trie() : tp(0) {}
+  virtual ~trie();		// virtual to shut up g++
+  void insert(const char *, int, void *);
+  // find calls do_match for each match it finds
+  void find(const char *pat, int patlen);
+  void clear();
+};
+
+class hyphen_trie : private trie {
+  int *h;
+  void do_match(int i, void *v);
+  void do_delete(void *v);
+  void insert_pattern(const char *pat, int patlen, int *num);
+public:
+  hyphen_trie() {}
+  ~hyphen_trie() {}
+  void hyphenate(const char *word, int len, int *hyphens);
+  void read_patterns_file(const char *name);
+};
+
+
+struct hyphenation_language {
+  symbol name;
+  dictionary exceptions;
+  hyphen_trie patterns;
+  hyphenation_language(symbol nm) : name(nm), exceptions(501) {}
+  ~hyphenation_language() { }
+};
+
+dictionary language_dictionary(5);
+hyphenation_language *current_language = 0;
+
+static void set_hyphenation_language()
+{
+  symbol nm = get_name(1);
+  if (!nm.is_null()) {
+    current_language = (hyphenation_language *)language_dictionary.lookup(nm);
+    if (!current_language) {
+      current_language = new hyphenation_language(nm);
+      (void)language_dictionary.lookup(nm, (void *)current_language);
+    }
+  }
+  skip_line();
+}
+
+const int WORD_MAX = 1024;
 
 static void hyphen_word()
 {
+  if (!current_language) {
+    error("no current hyphenation language");
+    skip_line();
+    return;
+  }
   char buf[WORD_MAX + 1];
   unsigned char pos[WORD_MAX + 2];
   for (;;) {
@@ -2757,7 +2799,8 @@ static void hyphen_word()
       buf[i] = 0;
       unsigned char *tem = new unsigned char[npos + 1];
       memcpy(tem, pos, npos+1);
-      tem = (unsigned char *)exception_dictionary.lookup(symbol(buf), tem);
+      tem = (unsigned char *)current_language->exceptions.lookup(symbol(buf),
+								 tem);
       if (tem)
 	a_delete tem;
     }
@@ -2765,26 +2808,12 @@ static void hyphen_word()
   skip_line();
 }
 
-struct trie_node;
-
-class trie {
-  trie_node *tp;
-  virtual void do_match(int len, void *val) = 0;
-public:
-  trie() : tp(0) {}
-  void insert(const char *, int, void *);
-  // find calls do_match for each match it finds
-  void find(const char *pat, int patlen);
-  void clear();
-};
-
 struct trie_node {
   char c;
   trie_node *down;
   trie_node *right;
   void *val;
   trie_node(char, trie_node *);
-  ~trie_node();
 };
 
 trie_node::trie_node(char ch, trie_node *p) 
@@ -2792,21 +2821,26 @@ trie_node::trie_node(char ch, trie_node *p)
 {
 }
 
-trie_node::~trie_node()
+trie::~trie()
 {
-  if (down)
-    delete down;
-  if (right)
-    delete right;
-  if (val)
-    delete val;
+  clear();
 }
 
 void trie::clear()
 {
-  if (tp) {
-    delete tp;
-    tp = 0;
+  delete_trie_node(tp);
+  tp = 0;
+}
+
+
+void trie::delete_trie_node(trie_node *p)
+{
+  if (p) {
+    delete_trie_node(p->down);
+    delete_trie_node(p->right);
+    if (p->val)
+      do_delete(p->val);
+    delete p;
   }
 }
 
@@ -2844,16 +2878,6 @@ void trie::find(const char *pat, int patlen)
   }
 }
 
-class hyphen_trie : private trie {
-  int *h;
-  void do_match(int i, void *v);
-  void insert_pattern(const char *pat, int patlen, int *num);
-public:
-  hyphen_trie() {}
-  void hyphenate(const char *word, int len, int *hyphens);
-  void read_patterns_file(const char *name);
-};
-
 struct operation {
   operation *next;
   short distance;
@@ -2899,21 +2923,40 @@ void hyphen_trie::do_match(int i, void *v)
   }
 }
 
-
+void hyphen_trie::do_delete(void *v)
+{
+  operation *op = (operation *)v;
+  while (op) {
+    operation *tem = op;
+    op = tem->next;
+    delete tem;
+  }
+}
+  
 void hyphen_trie::read_patterns_file(const char *name)
 {
   clear();
   char buf[WORD_MAX];
   int num[WORD_MAX+1];
   errno = 0;
-  FILE *fp = fopen(name, "r");
-  if (fp == 0)
-    fatal("can't open hyphenation patterns file `%1': %2",
-	  name, strerror(errno));
+  char *path = 0;
+  FILE *fp = macro_path.open_file(name, &path);
+  if (fp == 0) {
+    error("can't find hyphenation patterns file `%1'", name);
+    return;
+  }
   int c = getc(fp);
   for (;;) {
-    while (c != EOF && csspace(c))
+    for (;;) {
+      if (c == '%') {
+	do {
+	  c = getc(fp);
+	} while (c != EOF && c != '\n');
+      }
+      if (c == EOF || !csspace(c))
+	break;
       c = getc(fp);
+    }
     if (c == EOF)
       break;
     int i = 0;
@@ -2926,16 +2969,18 @@ void hyphen_trie::read_patterns_file(const char *name)
 	num[i] = 0;
       }
       c = getc(fp);
-    } while (i < WORD_MAX && c != EOF && !csspace(c));
+    } while (i < WORD_MAX && c != EOF && !csspace(c) && c != '%');
     insert_pattern(buf, i, num);
   }
   fclose(fp);
+  a_delete path;
+  return;
 }
 
-hyphen_trie ht;
-
 void hyphenate(hyphen_list *h, unsigned flags)
 {
+  if (!current_language)
+    return;
   while (h && h->hyphenation_code == 0)
     h = h->next;
   int len = 0;
@@ -2949,7 +2994,7 @@ void hyphenate(hyphen_list *h, unsigned flags)
   }
   if (len > 2) {
     buf[len] = 0;
-    unsigned char *pos = (unsigned char *)exception_dictionary.lookup(buf);
+    unsigned char *pos = (unsigned char *)current_language->exceptions.lookup(buf);
     if (pos != 0) {
       int j = 0;
       int i = 1;
@@ -2962,7 +3007,7 @@ void hyphenate(hyphen_list *h, unsigned flags)
     else {
       hbuf[0] = hbuf[len+1] = '.';
       int num[WORD_MAX+3];
-      ht.hyphenate(hbuf, len+2, num);
+      current_language->patterns.hyphenate(hbuf, len+2, num);
       int i;
       num[2] = 0;
       if (flags & 8)
@@ -2976,12 +3021,32 @@ void hyphenate(hyphen_list *h, unsigned flags)
   }
 }
 
-void read_hyphen_file(const char *name)
+static void hyphenation_patterns_file()
 {
-  ht.read_patterns_file(name);
+  symbol name = get_long_name(1);
+  if (!name.is_null()) {
+    if (!current_language)
+      error("no current hyphenation language");
+    else
+      current_language->patterns.read_patterns_file(name.contents());
+  }
+  skip_line();
+}
+
+class hyphenation_language_reg : public reg {
+public:
+  const char *get_string();
+};
+
+const char *hyphenation_language_reg::get_string()
+{
+  return current_language ? current_language->name.contents() : "";
 }
 
 void init_hyphen_requests()
 {
   init_request("hw", hyphen_word);
+  init_request("hla", set_hyphenation_language);
+  init_request("hpf", hyphenation_patterns_file);
+  number_reg_dictionary.define(".hla", new hyphenation_language_reg);
 }
